from __future__ import annotations

from typing import (
    Dict,
    List,
    Callable,
    Sequence,
    TYPE_CHECKING,
    Optional,
    Iterable,
)

import boto3
import botocore.exceptions

from access_undenied_aws import event
from access_undenied_aws import event_permission_data
from access_undenied_aws import logger

if TYPE_CHECKING:
    from mypy_boto3_iam.type_defs import ContextEntryTypeDef
else:
    ContextEntryTypeDef = object


class SimulateCustomPolicyContextGenerator(object):
    def __init__(
        self,
        session: boto3.Session,
        event_permission_data_: event_permission_data.EventPermissionData,
        cloudtrail_event_: event.Event,
    ):
        self.session = session
        self.iam_client = session.client("iam")
        self.event_permission_data = event_permission_data_
        self.cloudtrail_event = cloudtrail_event_

    def _get_aws_event_time(self) -> Optional[ContextEntryTypeDef]:
        return {
            "ContextKeyName": "aws:CurrentTime",
            "ContextKeyValues": (self.cloudtrail_event.event_time,),
            "ContextKeyType": "string",
        }

    def _get_aws_principal_arn(self) -> Optional[ContextEntryTypeDef]:
        return {
            "ContextKeyName": "aws:PrincipalArn",
            "ContextKeyValues": (self.event_permission_data.principal.arn,),
            "ContextKeyType": "string",
        }

    def _get_aws_principal_arn_caps(self) -> Optional[ContextEntryTypeDef]:
        context_entry = {
            "ContextKeyName": "aws:PrincipalARN",
            "ContextKeyValues": (self.event_permission_data.principal.arn,),
            "ContextKeyType": "string",
        }
        return context_entry

    def _get_aws_principal_tag(self, tag_key: str) -> Optional[ContextEntryTypeDef]:
        principal_tags = []
        try:
            if self.event_permission_data.principal.type == "AssumedRole":
                principal_tags = self.iam_client.list_role_tags(
                    RoleName=(self.event_permission_data.principal.name)
                )["Tags"]
            elif self.event_permission_data.principal.type == "IAMUser":
                principal_tags = self.iam_client.list_user_tags(
                    UserName=(self.event_permission_data.principal.name)
                )["Tags"]
        except botocore.exceptions.ClientError as list_tags_error:
            logger.error(
                f"[Error:{repr(list_tags_error)}] when getting" " aws:PrincipalTag value"
            )
            return None
        for tag in principal_tags:
            if tag["Key"] == tag_key:
                return {
                    "ContextKeyName": f"aws:PrincipalTag/{tag_key}",
                    "ContextKeyValues": (tag["Value"],),
                    "ContextKeyType": "string",
                }
        return None

    def _get_aws_requested_region(self) -> Optional[ContextEntryTypeDef]:
        if not self.cloudtrail_event.region:
            return None

        return {
            "ContextKeyName": "aws:RequestedRegion",
            "ContextKeyValues": (self.cloudtrail_event.region,),
            "ContextKeyType": "string",
        }

    def _get_aws_service_name(self) -> Optional[ContextEntryTypeDef]:
        if not self.cloudtrail_event.event_source:
            return None

        return {
            "ContextKeyName": "iam:AWSServiceName",
            "ContextKeyValues": (self.cloudtrail_event.event_source,),
            "ContextKeyType": "string",
        }

    def _get_aws_source_ip(self) -> Optional[ContextEntryTypeDef]:
        if not self.cloudtrail_event.source_ip_address:
            return None

        return {
            "ContextKeyName": "aws:SourceIp",
            "ContextKeyValues": (self.cloudtrail_event.source_ip_address,),
            "ContextKeyType": "string",
        }

    def _get_aws_source_vpce(self) -> Optional[ContextEntryTypeDef]:
        if not self.cloudtrail_event.vpc_endpoint_id:
            return None

        return {
            "ContextKeyName": "aws:sourceVpce",
            "ContextKeyValues": (self.cloudtrail_event.vpc_endpoint_id,),
            "ContextKeyType": "string",
        }

    def _get_aws_username(self) -> Optional[ContextEntryTypeDef]:
        return {
            "ContextKeyName": "aws:username",
            "ContextKeyValues": (self.event_permission_data.principal.name,),
            "ContextKeyType": "string",
        }

    def generate_context(
        self, context_keys: Iterable[str]
    ) -> Sequence[ContextEntryTypeDef]:
        context_entries = []
        for context_key in context_keys:
            context_generation_result = None
            if (
                context_key in SimulateCustomPolicyContextGenerator.KEY_FUNCTION_DICT
                or context_key
                in SimulateCustomPolicyContextGenerator.KEY_WITH_SUBKEY_FUNCTION_DICT
            ):
                context_generation_result = (
                    SimulateCustomPolicyContextGenerator.KEY_FUNCTION_DICT[context_key](
                        self
                    )
                )
            elif (
                context_key
                in SimulateCustomPolicyContextGenerator.KEY_WITH_SUBKEY_FUNCTION_DICT
                and "/" in context_key
            ):
                subkey = context_key.split("/", 1)[1]
                context_generation_result = (
                    SimulateCustomPolicyContextGenerator.KEY_WITH_SUBKEY_FUNCTION_DICT[
                        context_key
                    ](self, subkey)
                )
            if context_generation_result:
                context_entries.append(context_generation_result)
            else:
                logger.warning(
                    "Unable to find value for condition context key"
                    f" [context_key: {context_key}]"
                )
        return context_entries

    KEY_FUNCTION_DICT: Dict[
        str,
        Callable[
            [SimulateCustomPolicyContextGenerator],
            Optional[ContextEntryTypeDef],
        ],
    ] = {
        "aws:username": _get_aws_username,
        "aws:CurrentTime": _get_aws_event_time,
        "aws:PrincipalArn": _get_aws_principal_arn,
        "aws:PrincipalARN": _get_aws_principal_arn_caps,
        "aws:SourceVpce": _get_aws_source_vpce,
        "aws:SourceIp": _get_aws_source_ip,
        "aws:RequestedRegion": _get_aws_requested_region,
        "iam:AWSServiceName": _get_aws_service_name,
    }

    KEY_WITH_SUBKEY_FUNCTION_DICT: Dict[
        str,
        Callable[
            [SimulateCustomPolicyContextGenerator, str],
            Optional[ContextEntryTypeDef],
        ],
    ] = {
        "aws:PrincipalTag": _get_aws_principal_tag,
    }



from __future__ import annotations

import logging
from typing import IO

import boto3
import click
import click_log
import colorlog
import json

from access_undenied_aws import analysis
from access_undenied_aws import common
from access_undenied_aws import logger
from access_undenied_aws import organizations


def _initialize_logger() -> None:
    click_log.basic_config(logger)
    root_handler = logger.handlers[0]
    formatter = colorlog.ColoredFormatter(
        "%(log_color)s[%(asctime)s,%(msecs)d %(levelname)-8s"
        " %(filename)s:%(lineno)d - %(funcName)20s()]%(reset)s"
        " %(white)s%(message)s",
        datefmt="%H:%M:%S",
        reset=True,
        log_colors={
            "DEBUG": "blue",
            "INFO": "green",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "red",
        },
    )
    root_handler.setFormatter(formatter)


def initialize_config_from_user_input(
    config: common.Config,
    output_file: IO[str],
    management_account_role_arn: str,
    suppress_output: bool,
    cross_account_role_name: str,
) -> None:
    config.cross_account_role_name = cross_account_role_name
    config.management_account_role_arn = management_account_role_arn
    if logger.level == logging.NOTSET:
        logger.setLevel(logging.INFO)
    config.output_file = output_file
    config.suppress_output = suppress_output


_initialize_logger()
pass_config = click.make_pass_decorator(common.Config, ensure=True)


@click.group()
@click_log.simple_verbosity_option(logger)
@click.option(
    "--profile",
    help="the AWS profile to use (default is default profile)",
    default=None,
)
@pass_config
def access_undenied_aws(config: common.Config, profile: str) -> None:
    """
    Parses AWS AccessDenied CloudTrail events, explains the reasons for them, and offers actionable fixes.
    """
    config.session = boto3.Session(profile_name=profile)
    config.account_id = config.session.client("sts").get_caller_identity()["Account"]
    config.iam_client = config.session.client("iam")


@access_undenied_aws.command()
@click.option(
    "--events-file",
    help="input file of CloudTrail events",
    required=True,
    type=click.File("r"),
)
@click.option(
    "--scp-file",
    help="Service control policy data file generated by the get_scps command.",
    default=None,
    type=click.File("r"),
)
@click.option(
    "--management-account-role-arn",
    help=(
        "a cross-account role in the management account of the organization "
        "that must be assumable by your credentials."
    ),
    default=None,
)
@click.option(
    "--cross-account-role-name",
    help=(
        "The name of the cross-account role for AccessUndenied to assume."
        " default: AccessUndeniedRole"
    ),
    default="AccessUndeniedRole",
)
@click.option(
    "--output-file",
    help="output file for results (default: no output to file)",
    default=None,
    type=click.File("w"),
)
@click.option(
    "--suppress-output/--no-suppress-output",
    help="should output to stdout be suppressed (default: not suppressed)",
    default=False,
)
@pass_config
def analyze(
    config: common.Config,
    events_file: click.File,
    scp_file: IO[str],
    management_account_role_arn: str,
    cross_account_role_name: str,
    output_file: IO[str],
    suppress_output: bool,
) -> None:
    """
    Analyzes AWS CloudTrail events and explains the reasons for AccessDenied
    """
    initialize_config_from_user_input(
        config,
        output_file,
        management_account_role_arn,
        suppress_output,
        cross_account_role_name,
    )
    organizations.initialize_organization_data(config, scp_file.read())
    analysis.analyze_cloudtrail_events(config, events_file)


@access_undenied_aws.command()
@click.option(
    "--output-file",
    help="output file for scp data (default: scp_data.json)",
    default="scp_data.json",
    type=click.File("w"),
)
@pass_config
def get_scps(
    config: common.Config,
    output_file: IO[str],
) -> None:
    """
    Writes the organization's SCPs and organizational tree to a file
    """
    logger.info("Gathering Service Control Policy data...")
    organizations.initialize_organization_data(config, None)
    json.dump(config.organization_nodes, output_file, default=vars, indent=2)
    logger.info(f"Finished writing Service Control Policy data to {output_file.name}.")



from __future__ import annotations

import collections.abc
import re
from typing import Union, Tuple


def get_iterable(x: collections.Iterable) -> Union[object, Tuple[object]]:
    if x is None:
        return ()
    if isinstance(x, collections.abc.Iterable):
        return x
    else:
        return (x,)


def get_regex_match_group_or_none(match: re.Match, group_name: str):
    if not match:
        return None
    try:
        return match.group(group_name)
    except IndexError:
        return None



import boto3
import botocore
import cachetools
from cachetools import keys
import collections.abc
import json
from typing import List, TYPE_CHECKING, Optional, Union, Iterable

from access_undenied_aws import common
from access_undenied_aws import logger

if TYPE_CHECKING:
    from mypy_boto3_iam import IAMClient
    from mypy_boto3_iam.type_defs import (
        AttachedPolicyTypeDef,
        AttachedPermissionsBoundaryTypeDef,
        GetUserPolicyResponseTypeDef,
        GetGroupPolicyResponseTypeDef,
        GetRolePolicyResponseTypeDef,
    )
else:
    IAMClient = object
    AttachedPolicyTypeDef = object
    AttachedPermissionsBoundaryTypeDef = object
    GetUserPolicyResponseTypeDef = object
    GetGroupPolicyResponseTypeDef = object
    GetRolePolicyResponseTypeDef = object

InlinePolicyResponseTypeDef = Union[
    GetUserPolicyResponseTypeDef,
    GetGroupPolicyResponseTypeDef,
    GetRolePolicyResponseTypeDef,
]


def _create_policy_from_inline_policy_response(
    inline_policy_response: InlinePolicyResponseTypeDef,
    attachment_target_arn: str,
    attachment_target_type: common.AttachmentTargetType,
) -> common.Policy:
    # For some reason the get_(role|user|group)_policy API does not behave
    # according to the type hints, this line fixes that.
    inline_policy_document = (
        inline_policy_response["PolicyDocument"]
        if isinstance(inline_policy_response["PolicyDocument"], str)
        else json.dumps(inline_policy_response["PolicyDocument"])
    )
    return common.Policy(
        attachment_target_arn=attachment_target_arn,
        attachment_target_type=attachment_target_type,
        policy_name=inline_policy_response["PolicyName"],
        policy_arn="/".join(
            [attachment_target_arn, inline_policy_response["PolicyName"]]
        ),
        policy_document=inline_policy_document,
        policy_type=common.PolicyType.IDENTITY_INLINE_POLICY,
    )


def _create_policy_from_managed_policy_response(
    iam_client: IAMClient,
    managed_policy_response: AttachedPolicyTypeDef,
    attachment_target_arn: str,
    attachment_target_type: common.AttachmentTargetType,
) -> common.Policy:
    return common.Policy(
        attachment_target_arn=attachment_target_arn,
        attachment_target_type=attachment_target_type,
        policy_name=managed_policy_response["PolicyName"],
        policy_arn=managed_policy_response["PolicyArn"],
        policy_document=_get_policy_document_default_version_from_policy_arn(
            iam_client, managed_policy_response["PolicyArn"]
        ),
        policy_type=common.PolicyType.IDENTITY_MANAGED_POLICY,
    )


def _create_policy_from_permissions_boundary_response(
    iam_client: IAMClient,
    permission_boundary_response: AttachedPermissionsBoundaryTypeDef,
    attachment_target_arn: str,
    attachment_target_type: common.AttachmentTargetType,
) -> common.Policy:
    return common.Policy(
        attachment_target_arn=attachment_target_arn,
        attachment_target_type=attachment_target_type,
        policy_name=permission_boundary_response["PermissionsBoundaryArn"].split("/")[-1],
        policy_arn=permission_boundary_response["PermissionsBoundaryArn"],
        policy_document=_get_policy_document_default_version_from_policy_arn(
            iam_client,
            permission_boundary_response["PermissionsBoundaryArn"],
        ),
        policy_type=common.PolicyType.PERMISSIONS_BOUNDARY_POLICY,
    )


def _get_cross_account_iam_client(
    target_account: str, session: boto3.Session, cross_account_role_name: str
) -> IAMClient:
    try:
        role_credentials = session.client("sts").assume_role(
            RoleArn=(f"arn:aws:iam::{target_account}:role/{cross_account_role_name}"),
            RoleSessionName="AccessUndeniedResourceSession",
        )
        return boto3.Session(
            aws_access_key_id=role_credentials["Credentials"]["AccessKeyId"],
            aws_secret_access_key=role_credentials["Credentials"]["SecretAccessKey"],
            aws_session_token=role_credentials["Credentials"]["SessionToken"],
        ).client("iam")
    except botocore.exceptions.ClientError as client_error:
        msg = (
            "Could not assume cross-account role: [role_arn: "
            f"arn:aws:iam::{target_account}:role/{cross_account_role_name}]."
            f" [Error: {repr(client_error)}]."
        )
        logger.error(msg)
        raise common.AccessUndeniedError(msg, common.AccessDeniedReason.ERROR)


def _get_group_policies(
    iam_client: IAMClient, group_arn: str, group_name: str
) -> List[common.Policy]:
    group_policies = []
    for inline_policy_name in iam_client.list_group_policies(GroupName=group_name)[
        "PolicyNames"
    ]:
        group_policies.append(
            _create_policy_from_inline_policy_response(
                inline_policy_response=iam_client.get_group_policy(
                    GroupName=group_name, PolicyName=inline_policy_name
                ),
                attachment_target_arn=group_arn,
                attachment_target_type="IAMGroup",
            )
        )
    for managed_policy_response in iam_client.list_attached_group_policies(
        GroupName=group_name
    )["AttachedPolicies"]:
        group_policies.append(
            _create_policy_from_managed_policy_response(
                iam_client,
                managed_policy_response,
                group_arn,
                "IAMGroup",
            )
        )
    return group_policies


@cachetools.cached(
    cache=cachetools.LRUCache(maxsize=512),
    key=lambda iam_client, policy_arn: keys.hashkey(policy_arn),
)
def _get_policy_document_default_version_from_policy_arn(
    iam_client: IAMClient, policy_arn: str
) -> str:
    default_version = iam_client.get_policy(PolicyArn=policy_arn)["Policy"][
        "DefaultVersionId"
    ]
    policy_document = iam_client.get_policy_version(
        PolicyArn=policy_arn, VersionId=default_version
    )["PolicyVersion"]["Document"]
    return json.dumps(policy_document)


@cachetools.cached(
    cache=cachetools.LRUCache(maxsize=512),
    key=lambda iam_client, principal: keys.hashkey(principal.arn),
)
def _get_principal_inline_policies(
    iam_client: IAMClient, principal: common.Principal
) -> Iterable[common.Policy]:
    role_policies = []
    for inline_policy_name in iam_client.list_role_policies(RoleName=principal.name)[
        "PolicyNames"
    ]:
        role_policies.append(
            _create_policy_from_inline_policy_response(
                inline_policy_response=iam_client.get_role_policy(
                    RoleName=principal.name, PolicyName=inline_policy_name
                ),
                attachment_target_arn=principal.arn,
                attachment_target_type="AssumedRole",
            )
        )
    return role_policies


def _get_role_policies(
    iam_client: IAMClient, principal: common.Principal
) -> List[common.Policy]:
    role_policies = _get_principal_inline_policies(iam_client, principal)
    for managed_policy_response in iam_client.list_attached_role_policies(
        RoleName=principal.name
    )["AttachedPolicies"]:
        role_policies.append(
            _create_policy_from_managed_policy_response(
                iam_client=iam_client,
                managed_policy_response=managed_policy_response,
                attachment_target_arn=principal.name,
                attachment_target_type="AssumedRole",
            )
        )
    return role_policies


def _get_user_policies(
    iam_client: IAMClient, principal: common.Principal
) -> List[common.Policy]:
    user_policies = []
    for inline_policy_name in iam_client.list_user_policies(UserName=principal.name)[
        "PolicyNames"
    ]:
        user_policies.append(
            _create_policy_from_inline_policy_response(
                inline_policy_response=(
                    iam_client.get_user_policy(
                        UserName=principal.name, PolicyName=inline_policy_name
                    )
                ),
                attachment_target_arn=principal.arn,
                attachment_target_type="IAMUser",
            )
        )
    for managed_policy_response in iam_client.list_attached_user_policies(
        UserName=principal.name
    )["AttachedPolicies"]:
        user_policies.append(
            _create_policy_from_managed_policy_response(
                iam_client,
                managed_policy_response,
                principal.name,
                "IAMUser",
            )
        )
    return user_policies


def get_iam_client_in_account(
    config: common.Config,
    account_id: str,
) -> IAMClient:
    if config.account_id == account_id:
        return config.iam_client

    return _get_cross_account_iam_client(
        target_account=account_id,
        session=config.session,
        cross_account_role_name=config.cross_account_role_name,
    )


def get_iam_identity_policies_for_principal(
    iam_client: IAMClient, principal: common.Principal
) -> List[common.Policy]:
    logger.debug(f"Gathering identity policies for [principal_arn: {principal.arn}]")
    policies = []
    if principal.type == "AssumedRole":
        policies.extend(_get_role_policies(iam_client, principal))
    elif principal.type == "IAMUser":
        policies.extend(_get_user_policies(iam_client, principal))
        for group in iam_client.list_groups_for_user(UserName=(principal.name))["Groups"]:
            policies.extend(
                _get_group_policies(iam_client, group["Arn"], group["GroupName"])
            )
    return policies


def get_permissions_boundary_for_principal(
    iam_client: IAMClient, principal: common.Principal
) -> Optional[common.Policy]:
    principal_response = {}
    if principal.type == "IAMUser":
        principal_response = iam_client.get_user(UserName=(principal.name))["User"]
    elif principal.type == "AssumedRole":
        principal_response = iam_client.get_role(RoleName=(principal.name))["Role"]

    if principal_response.get("PermissionsBoundary"):
        return _create_policy_from_permissions_boundary_response(
            iam_client,
            iam_client.get_user(UserName=(principal.name))["User"]["PermissionsBoundary"],
            principal.arn,
            principal.type,
        )


def replace_principal_in_policy(
    original_principal: str, replacement_principal: str, policy: str
) -> str:
    policy_dict = json.loads(policy)
    for statement in policy_dict.get("Statement", []):
        principal_key = "Principal" if "Principal" in statement else "NotPrincipal"
        principal_value = statement.get(principal_key, {})
        if principal_value == original_principal:
            statement[principal_key] = replacement_principal
        elif isinstance(principal_value, collections.abc.Mapping) and principal_value.get("AWS") == original_principal:
            statement[principal_key]["AWS"] = replacement_principal
    return json.dumps(policy_dict)



import json
import re
from typing import Optional

import boto3
import botocore
from aws_error_utils import errors

from access_undenied_aws import common
from access_undenied_aws import event_permission_data
from access_undenied_aws import logger


def _get_ecr_resource_policy(
        arn_match: re.Match,
        session: boto3.Session,
        region: str,
        resource: common.Resource,
) -> Optional[common.Policy]:
    repository_policy_response = session.client(
        "ecr", region_name=region
    ).get_repository_policy(repositoryName=(arn_match.group("resource_id")))
    return common.Policy(
        attachment_target_arn=repository_policy_response["ARN"],
        attachment_target_type="Resource: ECR Repository",
        policy_name="ECRRepositoryResourcePolicy",
        policy_arn="/".join([resource.arn, "ECRRepositoryResourcePolicy"]),
        policy_document=repository_policy_response["ResourcePolicy"],
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def _get_iam_resource_policy(
        session: boto3.Session, resource: common.Resource
) -> Optional[common.Policy]:
    resource_policy_document = json.dumps(
        session.client("iam").get_role(RoleName=resource.arn.split("/")[-1])["Role"][
            "AssumeRolePolicyDocument"
        ]
    )
    return common.Policy(
        attachment_target_arn=resource.arn,
        attachment_target_type="Resource: IAM Role",
        policy_name="RoleTrustPolicy",
        policy_arn="/".join([resource.arn, "RoleTrustPolicy"]),
        policy_document=resource_policy_document,
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def _get_kms_resource_policy(
        arn_match: re.Match,
        session: boto3.Session,
        region: str,
        resource: common.Resource,
) -> Optional[common.Policy]:
    key_policy_document = session.client("kms", region_name=region).get_key_policy(
        KeyId=(arn_match.group("resource_id")), PolicyName="default"
    )["Policy"]
    return common.Policy(
        attachment_target_arn=resource.arn,
        attachment_target_type="Resource: KMS Key",
        policy_name="KMSKeyPolicy",
        policy_arn="/".join([resource.arn, "KMSKeyPolicy"]),
        policy_document=key_policy_document,
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def _get_lambda_resource_policy(
        arn_match: re.Match,
        session: boto3.Session,
        region: str,
        resource: common.Resource,
) -> Optional[common.Policy]:
    lambda_function_policy_response = session.client(
        "lambda", region_name=region
    ).get_policy(FunctionName=(arn_match.group("resource_id")))
    return common.Policy(
        attachment_target_arn=arn_match.group(0),
        attachment_target_type="Resource: Lambda Function",
        policy_name="LambdaFunctionResourcePolicy",
        policy_arn="/".join([resource.arn, "LambdaFunctionResourcePolicy"]),
        policy_document=lambda_function_policy_response["Policy"],
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def _get_resource_account_session(config: common.Config, resource: common.Resource) -> boto3.Session:
    if resource.account_id == config.account_id:
        return config.session

    role_arn = f"arn:aws:iam::{resource.account_id}:role/{config.cross_account_role_name}"
    try:
        role_credentials = config.session.client("sts").assume_role(
            RoleArn=role_arn, RoleSessionName="AccessUndeniedResourceSession"
        )["Credentials"]
        return boto3.Session(
            aws_access_key_id=role_credentials["AccessKeyId"],
            aws_secret_access_key=role_credentials["SecretAccessKey"],
            aws_session_token=role_credentials["SessionToken"],
        )
    except botocore.exceptions.ClientError as client_error:
        logger.error(
            f"Could not assume resource account role: {role_arn}:" f" {str(client_error)}"
        )
        raise common.AccessUndeniedError(
            f"[Error:{str(client_error)}] assuming [role_arn:{role_arn}] in the"
            f" resource account of [resource_arn={resource.arn}] when getting"
            " the resource policy.",
            common.AccessDeniedReason.ERROR,
        )


def _get_s3_resource_policy(
        arn_match: re.Match, session: boto3.Session, resource: common.Resource
) -> Optional[common.Policy]:
    bucket_name = arn_match.group("resource_type") or arn_match.group("resource_id")
    s3_client = session.client("s3")
    try:
        bucket_policy_document = s3_client.get_bucket_policy(Bucket=bucket_name)["Policy"]
    except errors.NoSuchBucketPolicy:
        bucket_policy_document = EMPTY_RESOURCE_POLICY.format(resource=resource.arn)
    return common.Policy(
        attachment_target_arn=resource.arn,
        attachment_target_type="Resource: S3 Bucket",
        policy_name="S3BucketPolicy",
        policy_arn="/".join([resource.arn, "S3BucketPolicy"]),
        policy_document=bucket_policy_document,
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def _get_secretsmanager_resource_policy(
        arn_match: re.Match,
        session: boto3.Session,
        region: str,
        resource: common.Resource,
) -> Optional[common.Policy]:
    secretsmanager_client = session.client("secretsmanager", region_name=region)
    secret_policy_response = secretsmanager_client.get_resource_policy(
        SecretId=(arn_match.group("resource_id"))
    )
    return common.Policy(
        attachment_target_arn=secret_policy_response["ARN"],
        attachment_target_type="Resource: SecretsManager Secret",
        policy_name="SecretResourcePolicy",
        policy_arn="/".join([resource.arn, "SecretResourcePolicy"]),
        policy_document=secret_policy_response["ResourcePolicy"],
        policy_type=common.PolicyType.RESOURCE_POLICY,
    )


def get_resource_policy(
        config: common.Config,
        event_permission_data_: event_permission_data.EventPermissionData,
        region: str,
) -> Optional[common.Policy]:
    if "*" in event_permission_data_.resource.arn:
        return None

    arn_match = re.search(
        common.RESOURCE_ARN_PATTERN,
        event_permission_data_.resource.arn,
        re.IGNORECASE,
    )
    if arn_match:
        service_name = arn_match.group("service")
    else:
        service_name = event_permission_data_.iam_permission.split(":")[0]
        if service_name != "secretsmanager":
            logger.warning(
                "Unable to parse service name from resource"
                f" [resource:{event_permission_data_.resource.arn}]"
                " ignoring resource policy..."
            )
            return None
    resource_account_session = _get_resource_account_session(config, event_permission_data_.resource)
    try:
        if service_name == "iam" and event_permission_data_.iam_permission in [
            "AssumeRole",
            "AssumeRoleWithSAML",
            "AssumeRoleWithWebIdentity",
        ]:
            return _get_iam_resource_policy(
                resource_account_session, event_permission_data_.resource
            )
        if service_name == "s3":
            return _get_s3_resource_policy(
                arn_match,
                resource_account_session,
                event_permission_data_.resource,
            )
        if service_name == "kms" and arn_match.group("resource_type") == "key":
            return _get_kms_resource_policy(
                arn_match,
                resource_account_session,
                region,
                event_permission_data_.resource,
            )
        if service_name == "secretsmanager":
            return _get_secretsmanager_resource_policy(
                arn_match,
                config.session,
                region,
                event_permission_data_.resource,
            )
        if service_name == "ecr":
            return _get_ecr_resource_policy(
                arn_match,
                config.session,
                region,
                event_permission_data_.resource,
            )
        if service_name == "lambda":
            return _get_lambda_resource_policy(
                arn_match,
                config.session,
                region,
                event_permission_data_.resource,
            )
    except botocore.exceptions.ClientError as client_error:
        raise common.AccessUndeniedError(
            f"[Error:{str(client_error)}] Getting resource policy for"
            f" [resource_arn={event_permission_data_.resource.arn}]",
            common.AccessDeniedReason.ERROR,
        )
    logger.warning(
        f"Service [service_name:{service_name}] does not have resource policy"
        " support in AccessUndenied, ignoring resource policy..."
    )
    return None


EMPTY_RESOURCE_POLICY = """
  {{
    "Version": "2012-10-17",
    "Statement": [
      {{
        "Effect": "Allow",
        "NotPrincipal": "*",
        "NotAction": "*",
        "Resource": "{resource}"
      }}
    ]
  }}
"""



from __future__ import annotations

from typing import Dict


class Event(object):
    def __init__(self, raw_event: Dict):
        self.error_code = raw_event.get("errorCode")
        self.error_message = raw_event.get("errorMessage")
        self.event_id = raw_event.get("eventID")
        self.event_name = raw_event.get("eventName")
        self.event_source = raw_event.get("eventSource")
        self.event_time = raw_event.get("eventTime")
        self.principal_type = raw_event.get("userIdentity", {}).get("type")
        self.raw_principal = raw_event.get("userIdentity")
        self.raw_resources = raw_event.get("resources")
        self.raw_request_parameters = raw_event.get("requestParameters")
        self.region = raw_event.get("awsRegion")
        self.source_ip_address = raw_event.get("sourceIPAddress")
        self.vpc_endpoint_id = raw_event.get("vpcEndpointId")



from __future__ import annotations

import json
from typing import Optional, List, Dict, Any

from access_undenied_aws import common, logger


class OrganizationNode(object):
    def __init__(
        self,
        arn: str,
        id_: str,
        name: str,
        organization_node_type: common.AttachmentTargetType,
        parent: Optional[str],
        policies: List[Dict[str, Any]],
    ):
        self.arn = arn
        self.id = id_
        self.name = name
        self.organization_node_type = organization_node_type
        self.policies = policies
        self.parent = parent
        self.combined_policy = self._create_combined_scp_policy()

    def _create_combined_scp_policy(self) -> str:
        statement_list = []
        for policy in self.policies:
            for statement in json.loads(policy["PolicyDocument"])["Statement"]:
                statement["Sid"] = (
                    f"{self.id}/{policy['Id']}/{policy['Name']}/{statement['Sid']}"
                    if "Sid" in statement
                    else f"{self.id}/{policy['Id']}/"
                )
                statement_list.append(statement)
        return json.dumps({"Version": "2012-10-17", "Statement": statement_list})


def create_scp_policies_from_organization_node(
    organization_nodes: Dict[str, OrganizationNode],
    account_id: str,
) -> List[common.Policy]:
    scp_policies = []
    if account_id not in organization_nodes:
        logger.warning(
            f"Could not read SCPs for [account_id:{account_id}], continuing"
            " without SCPs"
        )
        return []
    current_organization_node = organization_nodes[account_id]
    while current_organization_node:
        if current_organization_node.combined_policy:
            policy = common.Policy(
                attachment_target_arn=current_organization_node.id,
                attachment_target_type=current_organization_node.organization_node_type,
                policy_name=current_organization_node.id,
                policy_arn=current_organization_node.id,
                policy_document=current_organization_node.combined_policy,
                policy_type=common.PolicyType.COMBINED_SERVICE_CONTROL_POLICY,
            )
            scp_policies.append(policy)
        current_organization_node = organization_nodes.get(
            current_organization_node.parent
        )
    return scp_policies



from __future__ import annotations

import json
import re
from typing import Optional

import boto3
import botocore.exceptions
import pkg_resources

from access_undenied_aws import event
from access_undenied_aws import common
from access_undenied_aws import logger
from access_undenied_aws import utils

ACCESS_DENIED_MESSAGES = json.load(
    open(
        pkg_resources.resource_filename(
            __name__, "definitions/access_denied_message_patterns.json"
        ),
        "r",
    )
)

IAM_DATASET_MAPPINGS = json.load(
    open(pkg_resources.resource_filename(__name__, "definitions/map.json"), "r")
)


class EventPermissionData(object):
    def __init__(self) -> None:
        self.principal: common.Principal = common.Principal()
        self.iam_permission: str = ""
        self.resource: Optional[common.Resource] = None
        self.encoded_message: str = ""
        self.sdk_service_action: str = ""
        self.is_missing_allow_in_identity_based_policy = False

    @classmethod
    def from_event(
        cls,
        cloudtrail_event_: event.Event,
        config: common.Config,
    ) -> EventPermissionData:
        event_permission_data_ = cls()
        event_permission_data_.principal = _get_principal_from_user_identity(
            cloudtrail_event_, config
        )
        event_permission_data_._parse_error_message(cloudtrail_event_)
        if not event_permission_data_.iam_permission:
            event_permission_data_.iam_permission = (
                event_permission_data_._parse_iam_permission(cloudtrail_event_)
            )
        if not event_permission_data_.resource:
            event_permission_data_._parse_resource_from_cloudtrail_event(
                cloudtrail_event_
            )
        if not event_permission_data_.resource:
            event_permission_data_._parse_resource_from_iam_dataset_mapping(
                cloudtrail_event_
            )
        if not event_permission_data_.resource:
            event_permission_data_._parse_resource_from_common_request_parameters(
                cloudtrail_event_
            )
        if not event_permission_data_.resource:
            logger.debug(
                "Unable to parse resource field for"
                f" {cloudtrail_event_.event_source}:{cloudtrail_event_.event_name},"
                " setting '*' as default value."
            )
            event_permission_data_.resource = common.Resource(
                arn="*",
                account_id=event_permission_data_.principal.account_id,
            )
        return event_permission_data_

    def _parse_error_message(self, cloudtrail_event_: event.Event) -> None:
        if not cloudtrail_event_.error_message:
            return
        for access_denied_message_pattern in ACCESS_DENIED_MESSAGES[
            "invalidActionMessages"
        ]:
            if re.fullmatch(
                access_denied_message_pattern,
                cloudtrail_event_.error_message,
                re.IGNORECASE,
            ):
                raise common.AccessUndeniedError(
                    cloudtrail_event_.error_message,
                    common.AccessDeniedReason.INVALID_ACTION,
                )
        for access_denied_message_pattern in ACCESS_DENIED_MESSAGES[
            "accessDeniedMessages"
        ]:
            match = re.fullmatch(
                access_denied_message_pattern,
                cloudtrail_event_.error_message,
                re.IGNORECASE,
            )
            if not match:
                continue
            self.iam_permission = utils.get_regex_match_group_or_none(
                match, "iamPermission"
            )
            self._set_resource_parameter_from_error_message_match(match)
            self.is_missing_allow_in_identity_based_policy = (
                utils.get_regex_match_group_or_none(match, "missingAllowPolicyType")
                == "identity-based"
            )
            return
        logger.warning(
            "Unsupported AccessDenied message [ErrorMessage:"
            f" {cloudtrail_event_.error_message}]"
        )

    def _parse_iam_permission(self, cloudtrail_event_: event.Event) -> str:
        service_endpoint_name = cloudtrail_event_.event_source.split(".")[0]
        truncated_event_name = cloudtrail_event_.event_name.split("20")[0].split("V1")[0]
        sdk_service_names = IAM_DATASET_MAPPINGS["service_sdk_mappings"].get(
            service_endpoint_name
        ) or [service_endpoint_name]
        if service_endpoint_name not in IAM_DATASET_MAPPINGS["service_sdk_mappings"]:
            logger.warning(
                "Unidentified API service name [service_endpoint:"
                f" {service_endpoint_name}] from [event_source:"
                f" {cloudtrail_event_.event_source}], proceeding with the"
                f" service as [service_endpoint: {service_endpoint_name}]"
            )
        for sdk_service_name in sdk_service_names:
            sdk_service_action = f"{sdk_service_name}.{truncated_event_name}"
            if sdk_service_action in IAM_DATASET_MAPPINGS["sdk_permissionless_actions"]:
                raise common.AccessUndeniedError(
                    f"Permissionless action {sdk_service_action} cannot be" " denied.",
                    common.AccessDeniedReason.INVALID_ACTION,
                )
            if sdk_service_action in IAM_DATASET_MAPPINGS["sdk_method_iam_mappings"]:
                self.sdk_service_action = sdk_service_action
                action_from_mapping = IAM_DATASET_MAPPINGS["sdk_method_iam_mappings"][
                    sdk_service_action
                ][0]["action"]
                logger.debug(
                    "Got IAM permission from SDK->IAM mapping [iam_permission:"
                    f" {action_from_mapping}]..."
                )
                return action_from_mapping  # only the first permission

        logger.warning(
            "Unidentified API action"
            f" [sdk_service_names:{sdk_service_names}]:[action:{truncated_event_name}],"
            " proceeding with the IAM permission as"
            f" {service_endpoint_name}:{truncated_event_name}"
        )
        return f"{service_endpoint_name}:{truncated_event_name}"

    def _parse_resource_from_iam_dataset_mapping(
        self, cloudtrail_event_: event.Event
    ) -> None:
        if self.sdk_service_action not in IAM_DATASET_MAPPINGS["sdk_method_iam_mappings"]:
            return
        for iam_dataset_mapping_key, iam_dataset_mapping in IAM_DATASET_MAPPINGS[
            "sdk_method_iam_mappings"
        ][self.sdk_service_action][0]["resource_mappings"].items():
            logger.info(f"{iam_dataset_mapping_key}, {iam_dataset_mapping}")
            if "template" not in iam_dataset_mapping:
                continue
            template = iam_dataset_mapping["template"]
            # simple iam-dataset template: e.g. "template": "${RoleName}"
            resource_template_match = re.fullmatch(r"\${([^}]+)}", template)
            if not resource_template_match:
                continue
            if not cloudtrail_event_.raw_request_parameters:
                continue
            for (
                request_parameter,
                value,
            ) in cloudtrail_event_.raw_request_parameters.items():
                if request_parameter == resource_template_match.group(1):
                    if re.fullmatch(common.RESOURCE_ARN_PATTERN, value):
                        self.resource = common.Resource(
                            arn=value,
                            account_id=_parse_account_id_from_arn(
                                arn=value, default=self.principal.account_id
                            ),
                        )

    def _parse_resource_from_cloudtrail_event(
        self, cloudtrail_event_: event.Event
    ) -> None:
        for raw_resource in utils.get_iterable(cloudtrail_event_.raw_resources):
            for arn_key in ["arn", "Arn", "ARN"]:
                if arn_key in raw_resource:
                    # account id is the fourth element of arn:partition:service:region:account-id:resource
                    self.resource = common.Resource(
                        arn=raw_resource[arn_key],
                        account_id=(
                            raw_resource.get("accountId")
                            or _parse_account_id_from_arn(
                                arn=raw_resource[arn_key],
                                default=self.principal.account_id,
                            )
                        ),
                    )

    def _parse_resource_from_common_request_parameters(
        self, cloudtrail_event_: event.Event
    ) -> None:
        if cloudtrail_event_.raw_request_parameters:
            for resource_key in ["resource-arn", "id_"]:
                if resource_key not in cloudtrail_event_.raw_request_parameters:
                    continue
                resource_arn = cloudtrail_event_.raw_request_parameters[resource_key]
                self.resource = common.Resource(
                    arn=resource_arn,
                    account_id=_parse_account_id_from_arn(
                        arn=resource_arn, default=self.principal.account_id
                    ),
                )

    def _set_resource_parameter_from_error_message_match(
        self, error_message_regex_match: re.Match
    ) -> None:
        raw_resource_from_error_message = utils.get_regex_match_group_or_none(
            error_message_regex_match, "resource"
        )
        # ordinary resource arn with optional resource-type field
        # example 1: arn:aws:ec2:us-east-2:123456789012:instance/i-1234567812345678
        # example 2: arn:aws:s3:::my-bucket
        if raw_resource_from_error_message and re.match(
            r"^arn:[^:\n]*:[^:\n]*:[^:\n]*:[^:\n]*:(([^:/\n]*)[:/])?(.*)$",
            raw_resource_from_error_message,
        ):
            # account id is the fourth element of arn:partition:service:region:account-id:resource
            self.resource = common.Resource(
                raw_resource_from_error_message,
                _parse_account_id_from_arn(
                    raw_resource_from_error_message,
                    default=self.principal.account_id,
                ),
            )
            return

        raw_access_key_id_from_error_message = utils.get_regex_match_group_or_none(
            error_message_regex_match, "accessKeyId"
        )
        if raw_access_key_id_from_error_message:
            self.resource = common.Resource(
                arn=raw_access_key_id_from_error_message,
                account_id=self.principal.account_id,
            )
            return

        raw_role_name_from_error_message = utils.get_regex_match_group_or_none(
            error_message_regex_match, "roleName"
        )
        if raw_role_name_from_error_message:
            self.resource = common.Resource(
                arn=f"arn:aws:iam::{self.principal.account_id}:role/{raw_role_name_from_error_message}",
                account_id=self.principal.account_id,
            )
        return


def _get_principal_from_user_identity(
    event_: event.Event,
    config: common.Config,
) -> common.Principal:
    principal = common.Principal()
    principal.type = event_.principal_type
    if event_.principal_type == "IAMUser":
        principal.arn = principal.session_name = event_.raw_principal["arn"]
        principal.name = _get_principal_name_from_principal_arn(principal.arn)
        principal.account_id = _parse_account_id_from_arn(principal.arn)
        return principal

    if event_.principal_type == "AssumedRole":
        principal.arn = event_.raw_principal["sessionContext"]["sessionIssuer"]["arn"]
        principal.name = _get_principal_name_from_principal_arn(principal.arn)
        principal.session_name = event_.raw_principal["arn"]
        principal.account_id = _parse_account_id_from_arn(principal.arn)
        return principal

    if event_.principal_type == "AWSAccount":
        if event_.raw_principal["accountId"] == "ANONYMOUS_PRINCIPAL":
            raise common.AccessUndeniedError(
                "Anonymous principal access not yet supported.",
                common.AccessDeniedReason.ERROR,
            )
        # Handle the different behavior with a cross-account identity performing the action
        # in this case there is no principal ARN and we need to figure it out from the unique identifier
        # (AROAXXXX or AKIAXXX).
        # https://aws.amazon.com/blogs/security/aws-cloudtrail-now-tracks-cross-account-activity-to-its-origin/
        # Example:
        # "userIdentity": {
        #     "accountId": "123456789012",
        #     "principalId": "AROA6C17283581726ASDFAS:role-session-name",
        #     "type": "AWSAccount"
        # }
        logger.debug("Cross-account access: Principal represented as unique id...")
        principal.arn = _get_principal_arn_from_cross_account_principal_id(
            config.cross_account_role_name,
            event_.raw_principal["principalId"],
            event_.raw_principal["accountId"],
        )
        principal.type = "AssumedRole"
        principal.name = _get_principal_name_from_principal_arn(principal.arn)
        principal.session_name = principal.arn
        principal.account_id = event_.raw_principal["accountId"]
        return principal

    if event_.principal_type == "AWSService":
        raise common.AccessUndeniedError(
            "AWS Service actions not yet supported.",
            common.AccessDeniedReason.ERROR,
        )

    failed_to_analyze_raw_principal_message = (
        "Could not parse user identity data into a principal,"
        " this could be because analyzing root user actions"
        " is not yet supported. [userIdentityData:"
        f" {event_.raw_principal}."
    )
    logger.error(failed_to_analyze_raw_principal_message)
    raise common.AccessUndeniedError(
        failed_to_analyze_raw_principal_message,
        common.AccessDeniedReason.ERROR,
    )


def _get_principal_arn_from_cross_account_principal_id(
    config: common.Config,
    raw_principal_id: str,
    account_id: str,
) -> str:
    # Example: "AROA6C17283581726ASDFAS:role-session-name"
    principal_unique_id = raw_principal_id.split(":")[0]
    logger.debug(
        "Retrieving principal arn for cross-account for [principal_unique_id:"
        f" {principal_unique_id}]"
    )
    if config.account_id == account_id:
        iam_client = config.iam_client
    else:
        try:
            role_credentials = config.session.client("sts").assume_role(
                RoleArn=f"arn:aws:iam::{config.account_id}:role/{config.cross_account_role_name}",
                RoleSessionName=f"AccessUndeniedCrossAccount_{account_id}",
            )
            target_account_session = boto3.Session(
                aws_access_key_id=role_credentials["Credentials"]["AccessKeyId"],
                aws_secret_access_key=role_credentials["Credentials"]["SecretAccessKey"],
                aws_session_token=role_credentials["Credentials"]["SessionToken"],
            )
            iam_client = target_account_session.client("iam")
        except botocore.exceptions.ClientError:
            raise common.AccessUndeniedError(
                message=(
                    f"Could not assume role into account {account_id} while"
                    " parsing unique id"
                    f" [raw_principal_id={principal_unique_id}]."
                ),
                access_denied_reason=common.AccessDeniedReason.ERROR,
            )
    try:
        if principal_unique_id.startswith("AROA"):
            for page in iam_client.get_paginator("list_roles").paginate(
                PaginationConfig={"PageSize": 25}
            ):
                for role in page.get("Roles"):
                    if role["RoleId"] == principal_unique_id:
                        return role["Arn"]
        elif principal_unique_id.startswith("AIDA"):
            for page in iam_client.get_paginator("list_users").paginate(
                PaginationConfig={"PageSize": 25}
            ):
                for user in page.get("Users"):
                    if user["UserId"] == principal_unique_id:
                        return user["Arn"]
        else:
            raise common.AccessUndeniedError(
                f"Unknown unique id type for {principal_unique_id}",
                common.AccessDeniedReason.ERROR,
            )
    except botocore.exceptions.ClientError:
        raise common.AccessUndeniedError(
            "Could not list roles/users while parsing AROA in account" f" {account_id}",
            common.AccessDeniedReason.ERROR,
        )
    raise common.AccessUndeniedError(
        "Could not parse unique id into role arn",
        common.AccessDeniedReason.ERROR,
    )


def _get_principal_name_from_principal_arn(principal: str) -> str:
    # Regex for role/user/group arn, capturing only the principal name parameter
    principal_name_match = re.fullmatch(
        r"^arn:aws:iam::\d{12}:(?:user|role|group)/(?:[^\s/:]+/)*([^\s/:]+)$",
        principal,
        re.IGNORECASE,
    )
    if principal_name_match:
        return principal_name_match.group(1)
    raise ValueError(f"unable to retrieve name for target: {principal}")


def _parse_account_id_from_arn(arn: str, default="") -> str:
    # regex for matching account id from arn.
    # Example: arn:aws:kms:us-east-1:123456789012:[.....]
    account_id_match = re.fullmatch(
        r"arn:aws:[^:]*:[^:]*:(\d{12}|):[^\s]+", arn, re.IGNORECASE
    )
    if account_id_match:
        return account_id_match.group(1) or default
    raise common.AccessUndeniedError(
        f"Could not parse account number from [arn: {arn}]",
        common.AccessDeniedReason.ERROR,
    )



from aws_error_utils import errors
import boto3
import json
from typing import List, Dict, Any, TYPE_CHECKING, Optional, IO

from access_undenied_aws import common
from access_undenied_aws import logger
from access_undenied_aws import organization_node
from access_undenied_aws.organization_node import OrganizationNode

if TYPE_CHECKING:
    from mypy_boto3_organizations import OrganizationsClient
else:
    OrganizationsClient = object


def _deserialize_organization_nodes(
    organization_nodes_object: object,
) -> object:
    if (
        not isinstance(organization_nodes_object, dict)
        or "organization_node_type" not in organization_nodes_object
    ):
        return organization_nodes_object
    return OrganizationNode(
        arn=organization_nodes_object["arn"],
        id_=organization_nodes_object["id"],
        name=organization_nodes_object["name"],
        organization_node_type=organization_nodes_object["organization_node_type"],
        parent=organization_nodes_object["parent"],
        policies=organization_nodes_object["policies"],
    )


def _get_management_account_id(session: boto3.Session) -> Optional[str]:
    try:
        return (session.client("organizations").describe_organization())["Organization"][
            "MasterAccountId"
        ]
    except errors.AWSOrganizationsNotInUse:
        logger.debug("The account is not a member of an AWS Organization.")
        return None


def _get_management_account_organizations_client(
    session: boto3.Session,
    management_role_arn: str,
    management_account_id: str,
) -> Optional[OrganizationsClient]:
    if management_account_id == session.client("sts").get_caller_identity()["Account"]:
        logger.debug(f"The profile is in the organization's management account.")
        return session.client("organizations")

    if not management_role_arn:
        return None
    try:
        role_credentials = session.client("sts").assume_role(
            RoleArn=management_role_arn,
            RoleSessionName="AccessUndeniedManagementSession",
        )
        management_account_session = boto3.Session(
            aws_access_key_id=role_credentials["Credentials"]["AccessKeyId"],
            aws_secret_access_key=role_credentials["Credentials"]["SecretAccessKey"],
            aws_session_token=role_credentials["Credentials"]["SessionToken"],
        )
        return management_account_session.client("organizations")
    except errors.AccessDenied:
        logger.error(f"Could not assume organization role: {management_role_arn}")
        return None


def _get_organization_account_nodes(
    management_organizations_client,
) -> Dict[str, organization_node.OrganizationNode]:
    """
    SCPS at different levels of the hierarchy are treated as an AND condition (both must allow).
    Within one level of the hierarchy, different SCPs are treated as an OR (one allow is enough).
    See AWS documentation: Inheritance for service control policies
    https://docs.aws.amazon.com/organizations/latest/userguide/orgs_manage_policies_inheritance_auth.html
    """
    logger.debug("Traversing organization tree to retrieve all SCPs...")
    organization_root_response = management_organizations_client.list_roots()["Roots"][0]
    organization_nodes = {}
    logger.debug(
        "Analyzing SCPs for organization root"
        f" [organization_root: {organization_root_response['Id']}]"
    )
    organization_node_root = organization_node.OrganizationNode(
        arn=organization_root_response["Arn"],
        id_=organization_root_response["Id"],
        name=organization_root_response["Name"],
        organization_node_type="Organizational Unit or Organization Root",
        parent=None,
        policies=_get_target_policies_with_policy_document(
            organization_root_response["Id"], management_organizations_client
        ),
    )
    organization_nodes[organization_root_response["Id"]] = organization_node_root
    organizational_units_stack = [organization_node_root]
    while organizational_units_stack:  # while stack is not empty
        current_organizational_unit = organizational_units_stack.pop()
        for account in management_organizations_client.list_accounts_for_parent(
            ParentId=current_organizational_unit.id
        )["Accounts"]:
            logger.debug(f"Analyzing SCPs for [account_id: {account['Id']}]")
            organization_nodes[account["Id"]] = organization_node.OrganizationNode(
                arn=account["Arn"],
                id_=account["Id"],
                name=account["Name"],
                organization_node_type="AWSAccount",
                parent=current_organizational_unit.id,
                policies=_get_target_policies_with_policy_document(
                    account["Id"], management_organizations_client
                ),
            )

        for (
            organizational_unit_child_response
        ) in management_organizations_client.list_organizational_units_for_parent(
            ParentId=current_organizational_unit.id
        )[
            "OrganizationalUnits"
        ]:
            logger.debug(
                "Analyzing SCPs for [current_organizational_unit"
                f" {organizational_unit_child_response['Id']}]"
            )
            child_organizational_unit_node = organization_node.OrganizationNode(
                arn=organizational_unit_child_response["Arn"],
                id_=organizational_unit_child_response["Id"],
                name=organizational_unit_child_response["Name"],
                organization_node_type="Organizational Unit or Organization Root",
                parent=current_organizational_unit.id,
                policies=_get_target_policies_with_policy_document(
                    organizational_unit_child_response["Id"],
                    management_organizations_client,
                ),
            )
            organization_nodes[
                child_organizational_unit_node.id
            ] = child_organizational_unit_node
            organizational_units_stack.append(child_organizational_unit_node)

    return organization_nodes


def _get_target_policies_with_policy_document(
    target_id: str, management_organizations_client: OrganizationsClient
) -> List[Dict[str, Any]]:
    account_policies: List[
        Dict[str, Any]
    ] = management_organizations_client.list_policies_for_target(
        TargetId=target_id, Filter="SERVICE_CONTROL_POLICY"
    )[
        "Policies"
    ]
    for policy in account_policies:
        policy["PolicyDocument"] = management_organizations_client.describe_policy(
            PolicyId=policy["Id"]
        )["Policy"]["Content"]
    return account_policies


def initialize_organization_data(
    config: common.Config, scp_file_content: str
) -> None:
    config.management_account_id = _get_management_account_id(config.session)
    if scp_file_content:
        config.organization_nodes = json.loads(
            scp_file_content, object_hook=_deserialize_organization_nodes
        )
        logger.debug(
            "Organization data and SCPs successfully loaded from SCP data" " file..."
        )
        return

    if config.management_account_id:
        management_account_organizations_client = (
            _get_management_account_organizations_client(
                config.session,
                config.management_account_role_arn,
                config.management_account_id,
            )
        )
        if management_account_organizations_client:
            config.organization_nodes = _get_organization_account_nodes(
                management_account_organizations_client
            )



import json
from typing import (
    Dict,
    Any,
    Optional,
    TextIO,
    Union,
)

import botocore.exceptions
from aws_error_utils import errors

from access_undenied_aws import common
from access_undenied_aws import event
from access_undenied_aws import event_permission_data
from access_undenied_aws import iam_policy_data
from access_undenied_aws import logger
from access_undenied_aws import results
from access_undenied_aws import result_details
from access_undenied_aws import simulate_custom_policy_context_generator
from access_undenied_aws import simulate_custom_policy_helper
from access_undenied_aws.results import AnalysisResult


def _write_to_file(output_file: TextIO, output_json: Dict[str, Any]) -> None:
    if output_file is not None and output_file:
        output_file.write(json.dumps(output_json, indent=2))
        output_file.close()


def _handle_encoded_message(
    config: common.Config, error_message: str
) -> Optional[Dict[str, Any]]:
    # Handle the case of EC2 encoded message, decode the message and output the result
    # https://docs.aws.amazon.com/STS/latest/APIReference/API_DecodeAuthorizationMessage.html
    logger.debug("Encoded EC2 Authorization error message present, attempt to decode")
    return json.loads(
        config.session.client("sts").decode_authorization_message(
            EncodedMessage=error_message.split("Encoded authorization failure message: ")[
                -1
            ]
        )["DecodedMessage"]
    )


def _is_valid_encoded_message(error_message) -> bool:
    if error_message and "Encoded authorization failure message" in error_message:
        if error_message[-3:] == "...":
            logger.error(
                "Could not decode authorization message because the message"
                " in the CloudTrail event is incomplete. CloudTrail does"
                " not store error messages longer than 1024 chars."
                " Attempting regular analysis."
            )
            return False
        return True


def _write_to_output_buffer(
    config: common.Config,
    result: Union[Dict[str, Any], results.Result],
) -> None:
    """
    Output result to file, stdout, or both - as per the configuration.
    """
    if not config.suppress_output:
        print(str(result))
    if config.output_file:
        config.output_json.get("Results", []).append(json.loads(str(result)))


def analyze(config: common.Config, raw_event: Dict[str, Any]) -> Optional[results.Result]:
    if raw_event.get("errorCode") not in [
        "AccessDenied",
        "Client.UnauthorizedOperation",
    ]:
        logger.debug(
            "Event without AccessDenied/Client.UnauthorizedOperation"
            " error code not handled:"
            f" [eventID:{raw_event['eventID']}, "
            f"errorCode:{raw_event.get('errorCode')}]"
        )
        return None

    event_ = event.Event(raw_event)
    logger.info(f"AccessDenied cloudtrail_event found: [eventID={event_.event_id}]")
    logger.debug(f"[errorMessage:{event_.error_message}]")

    if _is_valid_encoded_message(event_.error_message):
        try:
            return _handle_encoded_message(config, event_.error_message)
        except (
            errors.InvalidAuthorizationMessageException,
        ) as invalid_authorization_message_exception:
            logger.debug(
                "Could not decode authorization message:"
                f" {invalid_authorization_message_exception}. Continuing"
                " analysis as usual."
            )

    try:
        logger.debug(
            "Analyzing event permission data (event Principal, Action,"
            " Resources, and Conditions)..."
        )
        event_permission_data_ = event_permission_data.EventPermissionData.from_event(
            event_, config
        )
        if event_permission_data_.is_missing_allow_in_identity_based_policy:
            logger.debug(
                "Quick-exit condition: simple missing allow in identity-based"
                " policy with all parameters known."
            )
            return _create_simple_missing_allow_analysis_result(
                cloudtrail_event_=event_,
                event_permission_data_=event_permission_data_,
            )

        logger.debug("Gathering all relevant IAM policies for permission analysis...")
        iam_policy_data_ = iam_policy_data.IamPolicyData.from_event_permission_data(
            config=config,
            event_permission_data_=event_permission_data_,
            region=event_.region,
        )

        logger.debug("Generating context values for condition assessment...")
        context = simulate_custom_policy_context_generator.SimulateCustomPolicyContextGenerator(
            session=config.session,
            event_permission_data_=event_permission_data_,
            cloudtrail_event_=event_,
        ).generate_context(
            simulate_custom_policy_helper.generate_context_key_list_for_simulate_custom_policy(
                iam_policy_data_, config.iam_client
            )
        )

        logger.debug(
            "Calling iam:SimulateCustomPolicy with the relevant arguments"
            " and analyzing the responses..."
        )
        simulate_custom_policy_request = (
            simulate_custom_policy_helper.generate_simulate_custom_policy_request(
                iam_policy_data_=iam_policy_data_,
                event_permission_data_=event_permission_data_,
                context=context,
            )
        )
        analysis_result = simulate_custom_policy_helper.simulate_custom_policies(
            config.iam_client,
            event_,
            event_permission_data_,
            iam_policy_data_,
            simulate_custom_policy_request,
        )
        if analysis_result:
            return analysis_result
    except common.AccessUndeniedError as invalid_authorization_message_exception:
        logger.error(
            str(invalid_authorization_message_exception)
            + f" [eventID:{raw_event.get('eventID', '<None>')}]"
        )
        return results.ErrorResult(
            event_id=raw_event.get("eventID", "<None>"),
            assessment_result=invalid_authorization_message_exception.access_denied_reason,
            error_message=str(invalid_authorization_message_exception),
        )

    except (
        errors.AccessDenied,
        botocore.exceptions.ClientError,
    ) as invalid_authorization_message_exception:
        logger.error(
            repr(invalid_authorization_message_exception)
            + f" [eventID:{raw_event.get('eventID', '<None>')}]"
        )
        return results.ErrorResult(
            event_id=raw_event.get("eventID", "<None>"),
            assessment_result=common.AccessDeniedReason.ERROR,
            error_message=str(invalid_authorization_message_exception),
        )


def analyze_cloudtrail_events(config: common.Config, raw_events_file_path):
    raw_events = json.load(raw_events_file_path)
    # Multiple CloudTrail log records in file or single event
    for raw_event in raw_events.get("Records") or [raw_events]:
        result = analyze(config, raw_event)
        if result:
            _write_to_output_buffer(config, result)
    _write_to_file(config.output_file, config.output_json)


def _create_simple_missing_allow_analysis_result(
    cloudtrail_event_: event.Event,
    event_permission_data_: event_permission_data.EventPermissionData,
) -> AnalysisResult:
    return AnalysisResult(
        event_id=cloudtrail_event_.event_id,
        assessment_result=common.AccessDeniedReason.IDENTITY_POLICY_MISSING_ALLOW,
        result_details_=result_details.MissingAllowResultDetails(
            simulate_custom_policy_response={
                "EvalActionName": event_permission_data_.iam_permission,
                "EvalResourceName": event_permission_data_.resource.arn,
                "EvalDecision": "implicitDeny",
                "MatchedStatements": [],
                "MissingContextValues": [],
            },
            missing_allow_reason=common.AccessDeniedReason.IDENTITY_POLICY_MISSING_ALLOW,
            target_resource_arn=event_permission_data_.resource.arn,
            target_principal_arn=event_permission_data_.principal.arn,
        ),
    )



import json
from typing import Optional, TYPE_CHECKING, Dict, Any

from access_undenied_aws import common, logger

if TYPE_CHECKING:
    from mypy_boto3_iam.type_defs import (
        EvaluationResultTypeDef,
    )
else:
    EvaluationResultTypeDef = object


class ResultDetails(object):
    def __init__(self) -> None:
        self.policies = []

    def __str__(self) -> str:
        return "{}"


class MissingAllowResultDetails(ResultDetails):
    def __init__(
        self,
        simulate_custom_policy_response: EvaluationResultTypeDef,
        missing_allow_reason: common.AccessDeniedReason,
        target_principal_arn: Optional[str] = None,
        target_resource_arn: Optional[str] = None,
    ) -> None:
        super(MissingAllowResultDetails, self).__init__()
        if missing_allow_reason in [
            common.AccessDeniedReason.IDENTITY_POLICY_MISSING_ALLOW,
            common.AccessDeniedReason.CROSS_ACCOUNT_MISSING_ALLOW,
            common.AccessDeniedReason.SCP_MISSING_ALLOW,
            common.AccessDeniedReason.PERMISSIONS_BOUNDARY_MISSING_ALLOW,
        ]:
            identity_policy = {
                "AttachmentTargetArn": target_principal_arn,
                "Policy": _create_new_identity_policy(simulate_custom_policy_response),
            }
            self.policies.append(identity_policy)
        if missing_allow_reason in [
            common.AccessDeniedReason.RESOURCE_POLICY_MISSING_ALLOW,
            common.AccessDeniedReason.CROSS_ACCOUNT_MISSING_ALLOW,
        ]:
            resource_policy = {
                "AttachmentTargetArn": target_resource_arn,
                "Policy": _create_new_resource_policy(
                    simulate_custom_policy_response, target_principal_arn
                ),
            }
            self.policies.append(resource_policy)

    def __str__(self) -> str:
        output_s = super(MissingAllowResultDetails, self).__str__()
        output = json.loads(output_s)
        logger.debug(f"[output:{repr(output)}]")
        output["PoliciesToAdd"] = self.policies
        return json.dumps(output, indent=2)


class ExplicitDenyResultDetails(ResultDetails):
    def __init__(
        self,
        policy_arn: str,
        policy_name: str,
        explicit_deny_policy_statement: str,
        attachment_target_arn,
    ) -> None:
        super(ExplicitDenyResultDetails, self).__init__()
        self.policies = [
            {
                "PolicyArn": policy_arn,
                "PolicyName": policy_name,
                "AttachmentTargetArn": attachment_target_arn,
                "PolicyStatement": json.loads(explicit_deny_policy_statement),
            }
        ]

    def __str__(self) -> str:
        output_s = super(ExplicitDenyResultDetails, self).__str__()
        output = json.loads(output_s)
        output["ExplicitDenyPolicies"] = self.policies
        return json.dumps(output, indent=2)


def _create_new_identity_policy(
    simulate_custom_policy_response: EvaluationResultTypeDef,
) -> Dict[str, Any]:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": simulate_custom_policy_response["EvalActionName"],
                "Resource": simulate_custom_policy_response["EvalResourceName"],
            }
        ],
    }


def _create_new_resource_policy(
    simulate_custom_policy_response: EvaluationResultTypeDef, principal_arn: str
) -> Dict[str, Any]:
    return {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {
                    "AWS": principal_arn,
                },
                "Action": simulate_custom_policy_response["EvalActionName"],
                "Resource": simulate_custom_policy_response["EvalResourceName"],
            }
        ],
    }



from __future__ import annotations

from access_undenied_aws import common
from access_undenied_aws import event_permission_data
from access_undenied_aws import iam_utils
from access_undenied_aws import logger
from access_undenied_aws import resource_policy_utils
from access_undenied_aws import organization_node


class IamPolicyData(object):
    def __init__(self) -> None:
        self.identity_policies = []
        self.guardrail_policies = []
        self.resource_policy = None
        self.caller_arn = ""
        self.caller_arn_placeholder = ""

    @classmethod
    def from_event_permission_data(
        cls,
        config: common.Config,
        event_permission_data_: event_permission_data.EventPermissionData,
        region: str,
    ) -> IamPolicyData:
        iam_policy_data = cls()
        iam_client = iam_utils.get_iam_client_in_account(
            config, event_permission_data_.principal.account_id
        )
        iam_policy_data.identity_policies = (
            iam_utils.get_iam_identity_policies_for_principal(
                iam_client, event_permission_data_.principal
            )
        )
        permissions_boundary = iam_utils.get_permissions_boundary_for_principal(
            iam_client, event_permission_data_.principal
        )
        if permissions_boundary:
            iam_policy_data.guardrail_policies.append(permissions_boundary)

        if event_permission_data_.principal.account_id == config.management_account_id:
            logger.debug(
                "The identity"
                f" [principal.arn:{event_permission_data_.principal.arn}] is in"
                " the management account, so it is not affected by SCPs."
            )
        else:
            iam_policy_data.guardrail_policies.extend(
                organization_node.create_scp_policies_from_organization_node(
                    config.organization_nodes,
                    event_permission_data_.principal.account_id,
                )
            )
        iam_policy_data.resource_policy = resource_policy_utils.get_resource_policy(
            config, event_permission_data_, region
        )
        iam_policy_data.caller_arn_placeholder = (
            event_permission_data_.principal.arn
            if event_permission_data_.principal.type == "IAMUser"
            else f"arn:aws:iam::{event_permission_data_.principal.account_id}:user/ZXJtZXRpY1JzY1RlYW0="
        )
        return iam_policy_data



import json
from typing import Optional, TYPE_CHECKING, Sequence, Dict, Any

from access_undenied_aws import (
    results,
    logger,
)
from access_undenied_aws import event
from access_undenied_aws import common
from access_undenied_aws import event_permission_data
from access_undenied_aws import iam_policy_data
from access_undenied_aws import result_details

if TYPE_CHECKING:
    from mypy_boto3_iam.type_defs import (
        EvaluationResultTypeDef,
        SimulateCustomPolicyRequestRequestTypeDef,
        StatementTypeDef,
    )
else:
    EvaluationResultTypeDef = object
    SimulateCustomPolicyRequestRequestTypeDef = object
    StatementTypeDef = object

POLICY_TYPE_TO_EXPLICIT_DENY_ASSESSMENT_RESULT_MAPPING = {
    common.PolicyType.IDENTITY_MANAGED_POLICY: common.AccessDeniedReason.IDENTITY_POLICY_EXPLICIT_DENY,
    common.PolicyType.IDENTITY_INLINE_POLICY: common.AccessDeniedReason.IDENTITY_POLICY_EXPLICIT_DENY,
    common.PolicyType.RESOURCE_POLICY: common.AccessDeniedReason.RESOURCE_POLICY_EXPLICIT_DENY,
    common.PolicyType.SERVICE_CONTROL_POLICY: common.AccessDeniedReason.SCP_EXPLICIT_DENY,
    common.PolicyType.PERMISSIONS_BOUNDARY_POLICY: common.AccessDeniedReason.PERMISSIONS_BOUNDARY_EXPLICIT_DENY,
    common.PolicyType.COMBINED_SERVICE_CONTROL_POLICY: common.AccessDeniedReason.SCP_EXPLICIT_DENY,
}


class SimulateCustomPolicyResultAnalyzer(object):
    def __init__(
        self,
        simulate_custom_policy_request: SimulateCustomPolicyRequestRequestTypeDef,
        simulate_custom_policy_response: EvaluationResultTypeDef,
        event_: event.Event,
        event_permission_data_: event_permission_data.EventPermissionData,
        iam_policy_data_: iam_policy_data.IamPolicyData,
        guardrail_policy: Optional[common.Policy],
    ) -> None:
        self.simulate_custom_policy_request = simulate_custom_policy_request
        self.simulate_custom_policy_response = simulate_custom_policy_response
        self.event = event_
        self.event_permission_data = event_permission_data_
        self.iam_policy_data_ = iam_policy_data_
        self.guardrail_policy = guardrail_policy

    def analyze(
        self,
    ) -> results.AnalysisResult:
        matched_explicit_deny_policy = _get_matched_explicit_deny_policy(
            simulate_custom_policy_request=self.simulate_custom_policy_request,
            simulate_custom_policy_response=self.simulate_custom_policy_response,
            event_permission_data_=self.event_permission_data,
            iam_policy_data_=self.iam_policy_data_,
            guardrail_policy=self.guardrail_policy,
        )
        if not matched_explicit_deny_policy:
            missing_allow_reason = self._identify_missing_allow_reason()
            return results.AnalysisResult(
                event_id=self.event.event_id,
                assessment_result=missing_allow_reason,
                result_details_=self._create_new_result_details(missing_allow_reason),
            )

        if (
            matched_explicit_deny_policy.policy_type
            is common.PolicyType.COMBINED_SERVICE_CONTROL_POLICY
        ):
            matched_explicit_deny_policy = _get_matched_scp_from_combined_scp(
                matched_explicit_deny_policy
            )

        return results.AnalysisResult(
            event_id=self.event.event_id,
            assessment_result=POLICY_TYPE_TO_EXPLICIT_DENY_ASSESSMENT_RESULT_MAPPING[
                matched_explicit_deny_policy.policy_type
            ],
            result_details_=result_details.ExplicitDenyResultDetails(
                policy_arn=matched_explicit_deny_policy.policy_arn,
                policy_name=matched_explicit_deny_policy.policy_name,
                explicit_deny_policy_statement=matched_explicit_deny_policy.matched_statement,
                attachment_target_arn=matched_explicit_deny_policy.attachment_target_arn,
            ),
        )

    def _create_new_result_details(
        self, missing_allow_reason: common.AccessDeniedReason
    ) -> result_details.ResultDetails:
        target_principal_arn = (
            self.guardrail_policy.attachment_target_arn
            if missing_allow_reason
            in [
                common.AccessDeniedReason.SCP_MISSING_ALLOW,
                common.AccessDeniedReason.PERMISSIONS_BOUNDARY_MISSING_ALLOW,
            ]
            else self.event_permission_data.principal.arn
        )
        return result_details.MissingAllowResultDetails(
            simulate_custom_policy_response=self.simulate_custom_policy_response,
            missing_allow_reason=missing_allow_reason,
            target_principal_arn=target_principal_arn,
            target_resource_arn=self.event_permission_data.resource.arn,
        )

    def _identify_missing_allow_reason(self) -> common.AccessDeniedReason:
        if self.simulate_custom_policy_response["EvalDecision"] == "allowed":
            return common.AccessDeniedReason.ALLOWED

        # EvalDecisionDetails only appears in cross-account-cases (or IAM Role/KMS Key)
        cross_account_evaluation_decision_details = (
            self.simulate_custom_policy_response.get("EvalDecisionDetails", {})
        )
        resource_policy_cross_account_implicit_deny = (
            cross_account_evaluation_decision_details.get("Resource Policy")
            == "implicitDeny"
        )
        identity_policy_cross_account_implicit_deny = (
            cross_account_evaluation_decision_details.get("IAM Policy") == "implicitDeny"
        )
        if identity_policy_cross_account_implicit_deny:
            if resource_policy_cross_account_implicit_deny:
                return common.AccessDeniedReason.CROSS_ACCOUNT_MISSING_ALLOW
            return common.AccessDeniedReason.IDENTITY_POLICY_MISSING_ALLOW

        if resource_policy_cross_account_implicit_deny:
            return common.AccessDeniedReason.RESOURCE_POLICY_MISSING_ALLOW

        if (
            cross_account_evaluation_decision_details.get("Permissions Boundary Policy")
            == "implicitDeny"
        ):
            if self.guardrail_policy.policy_type == "Permissions Boundary Policy":
                return common.AccessDeniedReason.PERMISSIONS_BOUNDARY_MISSING_ALLOW
            return common.AccessDeniedReason.SCP_MISSING_ALLOW

        # same-account case
        if (
            self.simulate_custom_policy_response.get(
                "PermissionsBoundaryDecisionDetail", {}
            ).get("PermissionsBoundaryDecisionDetail")
            is False
        ):
            return (
                common.AccessDeniedReason.PERMISSIONS_BOUNDARY_MISSING_ALLOW
                if self.guardrail_policy.policy_type
                is common.PolicyType.PERMISSIONS_BOUNDARY_POLICY
                else common.AccessDeniedReason.SCP_MISSING_ALLOW
            )

        return common.AccessDeniedReason.IDENTITY_POLICY_MISSING_ALLOW


def _extract_matching_statement(
    matched_statement_response: StatementTypeDef, policy_lines: Sequence[str]
) -> Optional[Dict[str, Any]]:
    line_start_index = matched_statement_response["StartPosition"]["Line"] - 1
    line_end_index = matched_statement_response["EndPosition"]["Line"] - 1
    if line_start_index != line_end_index:
        # Every single SimulateCustomPolicy response we have seen has returned
        # a single-line match and ignored line breaks in the original policy
        raise common.AccessUndeniedError(
            "SimulateCustomPolicy returned multiline statement match,"
            "currently unsupported.",
            common.AccessDeniedReason.ERROR,
        )

    col_start_index = matched_statement_response["StartPosition"]["Column"] - 1
    col_end_index = matched_statement_response["EndPosition"]["Column"] - 1
    try:
        return json.loads(
            policy_lines[line_start_index][col_start_index:col_end_index].lstrip(", ")
        )

    except json.JSONDecodeError:
        error_message = (
            "Could not decode matching statement"
            f" [matched_statement_response:{matched_statement_response}],"
            f" [policy_document:{str(policy_lines)}]"
        )
        logger.error(error_message)
        raise common.AccessUndeniedError(
            error_message,
            common.AccessDeniedReason.ERROR,
        )


def _get_matched_explicit_deny_policy(
    simulate_custom_policy_request: SimulateCustomPolicyRequestRequestTypeDef,
    simulate_custom_policy_response: EvaluationResultTypeDef,
    event_permission_data_: event_permission_data.EventPermissionData,
    iam_policy_data_: iam_policy_data.IamPolicyData,
    guardrail_policy: common.Policy,
) -> Optional[common.MatchedPolicy]:
    if not simulate_custom_policy_response.get(
        "MatchedStatements"
    ) and not simulate_custom_policy_response.get("ResourceSpecificResults", [{}])[0].get(
        "MatchedStatements"
    ):
        return None

    for matched_statement in simulate_custom_policy_response.get(
        "ResourceSpecificResults", [{}]
    )[0].get("MatchedStatements") or simulate_custom_policy_response.get(
        "MatchedStatements"
    ):
        policy = None
        policy_document_lines = []
        policy_type = matched_statement["SourcePolicyType"]
        if policy_type == "IAM Policy":
            policy_ind = int(matched_statement["SourcePolicyId"].split(".")[-1]) - 1
            policy = iam_policy_data_.identity_policies[policy_ind]
            policy_document_lines = simulate_custom_policy_request["PolicyInputList"][
                policy_ind
            ].splitlines()
        if policy_type == "Permissions Boundary Policy":
            policy = guardrail_policy
            policy_document_lines = simulate_custom_policy_request[
                "PermissionsBoundaryPolicyInputList"
            ][0].splitlines()
        if policy_type == "Resource Policy":
            policy = iam_policy_data_.resource_policy
            policy_document_lines = simulate_custom_policy_request[
                "ResourcePolicy"
            ].splitlines()

        matched_statement = _extract_matching_statement(
            matched_statement, policy_document_lines
        )
        if matched_statement.get("Effect") != "Deny":
            continue

        if policy_type == "Resource Policy":
            if (
                iam_policy_data_.caller_arn_placeholder
                != event_permission_data_.principal.arn
            ):
                # Replace caller arn placeholder with original caller arn
                matched_statement = json.loads(
                    json.dumps(matched_statement).replace(
                        iam_policy_data_.caller_arn_placeholder,
                        iam_policy_data_.caller_arn,
                    )
                )
        return common.MatchedPolicy(json.dumps(matched_statement), policy)
    return None


def _get_matched_scp_from_combined_scp(
    matched_policy: common.MatchedPolicy,
) -> common.MatchedPolicy:
    statement = json.loads(matched_policy.matched_statement)
    if "Sid" not in statement or len(statement["Sid"].split("/")) != 4:
        raise common.AccessUndeniedError(
            "Could not get matched SCP from combined SCP, "
            f"[matched_policy: {matched_policy}].",
            common.AccessDeniedReason.ERROR,
        )

    _, policy_id, policy_name, statement_sid = statement["Sid"].split("/")
    statement["Sid"] = statement_sid
    return common.MatchedPolicy(
        matched_statement=json.dumps(statement),
        policy=common.Policy(
            attachment_target_arn=matched_policy.attachment_target_arn,
            attachment_target_type=matched_policy.attachment_target_type,
            policy_name=policy_name,
            policy_arn=policy_id,
            policy_document=json.dumps(statement),
            policy_type=common.PolicyType.SERVICE_CONTROL_POLICY,
        ),
    )



from __future__ import annotations

import json

from access_undenied_aws import common
from access_undenied_aws import result_details


class Result(object):
    def __init__(
        self, event_id: str, assessment_result: common.AccessDeniedReason
    ) -> None:
        self.event_id = event_id
        self.assessment_result = assessment_result

    def __str__(self) -> str:
        output = dict()
        output["EventId"] = self.event_id
        if self.assessment_result:
            output["AssessmentResult"] = self.assessment_result
        return json.dumps(output, indent=2)


class ErrorResult(Result):
    def __init__(
        self,
        event_id: str,
        assessment_result: common.AccessDeniedReason,
        error_message: str,
    ):
        super(ErrorResult, self).__init__(event_id, assessment_result)
        self.error_message = error_message

    def __str__(self) -> str:
        output = json.loads(super(ErrorResult, self).__str__())
        output["ErrorMessage"] = self.error_message
        return json.dumps(output, indent=2)


class AnalysisResult(Result):
    def __init__(
        self,
        event_id: str,
        assessment_result: common.AccessDeniedReason,
        result_details_: result_details.ResultDetails,
    ):
        super().__init__(event_id, assessment_result)
        self.result_details = result_details_

    def __str__(self) -> str:
        output = json.loads(super(AnalysisResult, self).__str__())
        if self.result_details:
            output["ResultDetails"] = json.loads(str(self.result_details))
        return json.dumps(output, indent=2)



from __future__ import annotations

import enum
from typing import Literal

RESOURCE_ARN_PATTERN = (
    r"^arn:(?P<partition>[^:\n]*):(?P<service>[^:\n]*):(?P<region>[^:\n]*):"
    r"(?P<account_id>[^:\n]*):(?P<ignore>(?P<resource_type>[^:\/\n]*)[:\/])?(?P<resource_id>.*)$"
)

AttachmentTargetType = Literal[
    "AWSAccount",
    "IAMGroup",
    "AssumedRole",
    "IAMUser",
    "Organizational Unit or Organization Root",
    "Resource: ECR Repository",
    "Resource: IAM Role",
    "Resource: KMS Key",
    "Resource: Lambda Function",
    "Resource: S3 Bucket",
    "Resource: SecretsManager Secret",
]


class AccessDeniedReason(str, enum.Enum):
    ALLOWED = "Allowed"
    CROSS_ACCOUNT_MISSING_ALLOW = (
        "Missing allows in an identity-based and "
        "a resource-based policy in cross-account access"
    )
    IDENTITY_POLICY_EXPLICIT_DENY = "Explicit deny in an identity-based policy"
    IDENTITY_POLICY_MISSING_ALLOW = "Missing allow in an identity-based policy"
    PERMISSIONS_BOUNDARY_EXPLICIT_DENY = "Explicit deny in a permissions boundary policy"
    PERMISSIONS_BOUNDARY_MISSING_ALLOW = "Missing allow in a permissions boundary policy"
    RESOURCE_POLICY_EXPLICIT_DENY = "Explicit deny in a resource-based policy"
    RESOURCE_POLICY_MISSING_ALLOW = (
        "Missing allow in a resource-based policy in cross-account access"
    )
    SCP_EXPLICIT_DENY = "Explicit deny in a service-control policy"
    SCP_MISSING_ALLOW = "Missing allow in a service-control policy"
    INVALID_ACTION = "Invalid action"
    ERROR = "Error"


class PolicyType(str, enum.Enum):
    IDENTITY_INLINE_POLICY = "Inline IAM Policy"
    IDENTITY_MANAGED_POLICY = "Managed IAM Policy"
    RESOURCE_POLICY = "Resource Policy"
    PERMISSIONS_BOUNDARY_POLICY = "Permissions Boundary Policy"
    SERVICE_CONTROL_POLICY = "Service Control Policy"
    COMBINED_SERVICE_CONTROL_POLICY = "Merged Composite Service Control Policy"


class AccessUndeniedError(Exception):
    def __init__(self, message: str, access_denied_reason: AccessDeniedReason):
        super(AccessUndeniedError, self).__init__(message)
        self.message = message
        self.access_denied_reason = access_denied_reason


class Config(object):
    def __init__(self) -> None:
        self.cross_account_role_name = None
        self.management_account_role_arn = None
        self.management_account_id = ""
        self.organization_nodes = dict()
        self.output_file = None
        self.output_json = {"Results": []}
        self.session = None
        self.suppress_output = False
        self.account_id = None
        self.iam_client = None


class Policy(object):
    def __init__(
        self,
        attachment_target_arn: str,
        attachment_target_type: AttachmentTargetType,
        policy_name: str,
        policy_arn: str,
        policy_document: str,
        policy_type: PolicyType,
    ) -> None:
        self.attachment_target_arn = attachment_target_arn
        self.attachment_target_type = attachment_target_type
        self.policy_name = policy_name
        self.policy_arn = policy_arn
        self.policy_document = policy_document
        self.policy_type = policy_type


class MatchedPolicy(Policy):
    def __init__(self, matched_statement: str, policy: Policy):
        super().__init__(
            policy.attachment_target_arn,
            policy.attachment_target_type,
            policy.policy_name,
            policy.policy_arn,
            policy.policy_document,
            policy.policy_type,
        )
        self.matched_statement = matched_statement


class Principal(object):
    def __init__(self) -> None:
        self.session_name = ""
        self.arn = ""
        self.type = ""
        self.account_id = ""
        self.name = ""


class Resource(object):
    def __init__(self, arn: str, account_id: str) -> None:
        self.arn = arn
        self.account_id = account_id



import copy
import json
import re
from typing import (
    List,
    Sequence,
    Optional,
    TYPE_CHECKING,
    Dict,
    Any,
    Set,
    Iterable,
)

from access_undenied_aws import simulate_custom_policy_result_analyzer
from access_undenied_aws import (
    event_permission_data,
    iam_utils,
    event,
    iam_policy_data,
    common,
    results,
    utils,
    logger,
)
from access_undenied_aws.iam_policy_data import IamPolicyData

if TYPE_CHECKING:
    from mypy_boto3_iam import IAMClient
    from mypy_boto3_iam.type_defs import (
        ContextEntryTypeDef,
        SimulateCustomPolicyRequestRequestTypeDef,
    )
else:
    ContextEntryTypeDef = object
    IAMClient = object
    SimulateCustomPolicyRequestRequestTypeDef = object


def _add_resource_field(resource_policy, resource_arn) -> str:
    """
    Policy simulator requires resource policies to have the
    resource field explicitly stated. That is not the case for all resource
    policies (e.g. IAM Trust Policies)
    """
    resource_policy_dict = json.loads(resource_policy)
    for statement in resource_policy_dict.get("Statement", []):
        statement.pop("NotResource", None)
        statement["Resource"] = resource_arn
    return json.dumps(resource_policy_dict)


def _get_context_keys_for_custom_policy(
    policy_input_list: List[str],
) -> Set[str]:
    context_keys = set()
    for policy_document in policy_input_list:
        for statement in json.loads(policy_document).get("Statement", []):
            for _, condition_type_map in statement.get("Condition", {}).items():
                for context_key in condition_type_map.keys():
                    context_keys.add(context_key)
    return context_keys


def _simulate_custom_policy(
    iam_client: IAMClient,
    cloudtrail_event_: event.Event,
    event_permission_data_: event_permission_data.EventPermissionData,
    iam_policy_data_: iam_policy_data.IamPolicyData,
    guardrail_policy: Optional[common.Policy],
    simulate_custom_policy_arguments_template: SimulateCustomPolicyRequestRequestTypeDef,
) -> Optional[results.AnalysisResult]:
    if guardrail_policy:
        simulate_custom_policy_arguments = copy.copy(
            simulate_custom_policy_arguments_template
        )
        simulate_custom_policy_arguments["PermissionsBoundaryPolicyInputList"] = [
            guardrail_policy.policy_document
        ]
    else:
        simulate_custom_policy_arguments = simulate_custom_policy_arguments_template

    simulate_custom_policy_response = iam_client.simulate_custom_policy(
        **simulate_custom_policy_arguments
    )["EvaluationResults"][0]
    if simulate_custom_policy_response["EvalDecision"] in [
        "explicitDeny",
        "implicitDeny",
    ]:
        return simulate_custom_policy_result_analyzer.SimulateCustomPolicyResultAnalyzer(
            simulate_custom_policy_request=simulate_custom_policy_arguments,
            simulate_custom_policy_response=simulate_custom_policy_response,
            event_=cloudtrail_event_,
            event_permission_data_=event_permission_data_,
            iam_policy_data_=iam_policy_data_,
            guardrail_policy=guardrail_policy,
        ).analyze()
    return None


def generate_context_key_list_for_simulate_custom_policy(
    iam_policy_data_: IamPolicyData, iam_client: IAMClient
) -> Iterable[str]:
    policy_input_list = [
        identity_policy.policy_document
        for identity_policy in iam_policy_data_.identity_policies
    ] + [
        boundary_policy.policy_document
        for boundary_policy in iam_policy_data_.guardrail_policies
    ]
    if iam_policy_data_.resource_policy:
        policy_input_list.append(iam_policy_data_.resource_policy.policy_document)
    return _get_context_keys_for_custom_policy(policy_input_list)


def generate_simulate_custom_policy_request(
    iam_policy_data_: IamPolicyData,
    event_permission_data_: event_permission_data.EventPermissionData,
    context: Sequence[ContextEntryTypeDef],
) -> SimulateCustomPolicyRequestRequestTypeDef:
    simulate_custom_policy_request = {
        "PolicyInputList": [
            policy.policy_document for policy in iam_policy_data_.identity_policies
        ],
        "ActionNames": (event_permission_data_.iam_permission,),
        "ResourceOwner": get_resource_owner_parameter_from_account_arn(
            resource_arn=event_permission_data_.resource.arn,
            resource_account_id=event_permission_data_.resource.account_id,
            iam_permission=event_permission_data_.iam_permission,
        ),
        "CallerArn": iam_policy_data_.caller_arn_placeholder,
        "ResourceArns": (event_permission_data_.resource.arn,),
        "ContextEntries": context,
    }

    if iam_policy_data_.resource_policy:
        # We can only perform one principal replacement,
        # but the principal parameter in the resource policy can be
        # arn:aws:iam::account:role/role-name or it can be
        # arn:aws:sts::account:assumed-role/role-name/role-session-name
        # We need to find out which principal, if any,
        # is used in the resource policy. :(
        iam_policy_data_.caller_arn = (
            event_permission_data_.principal.session_name
            if event_permission_data_.principal.session_name
            in iam_policy_data_.resource_policy.policy_document
            else event_permission_data_.principal.arn
        )
        simulate_custom_policy_request[
            "ResourcePolicy"
        ] = iam_utils.replace_principal_in_policy(
            original_principal=iam_policy_data_.caller_arn,
            replacement_principal=iam_policy_data_.caller_arn_placeholder,
            policy=_add_resource_field(
                iam_policy_data_.resource_policy.policy_document,
                event_permission_data_.resource.arn,
            ),
        )
    return simulate_custom_policy_request


def get_resource_owner_parameter_from_account_arn(
    resource_arn: str,
    resource_account_id: str,
    iam_permission: str,
) -> str:
    arn_match = re.match(common.RESOURCE_ARN_PATTERN, resource_arn)
    if (
        utils.get_regex_match_group_or_none(arn_match, "resource_type") == "key"
        or "AssumeRole" in iam_permission
    ):
        logger.debug(
            "IAM Role trust policies and KMS Key Policies are"
            " anomalous and are evaluated like cross-account"
            " policies."
            " Listing placeholder account id 123456789012..."
        )
        return f"arn:aws:iam::123456789012:root"

    return f"arn:aws:iam::{resource_account_id}:root"


def simulate_custom_policies(
    iam_client: IAMClient,
    cloudtrail_event_: event.Event,
    event_permission_data_: event_permission_data.EventPermissionData,
    iam_policy_data_: iam_policy_data.IamPolicyData,
    simulate_custom_policy_arguments_base: SimulateCustomPolicyRequestRequestTypeDef,
) -> Optional[results.AnalysisResult]:
    for guardrail_policy in iam_policy_data_.guardrail_policies or [None]:
        deny_result = _simulate_custom_policy(
            iam_client,
            cloudtrail_event_,
            event_permission_data_,
            iam_policy_data_,
            guardrail_policy,
            simulate_custom_policy_arguments_base,
        )
        if deny_result:
            return deny_result
    raise common.AccessUndeniedError(
        message="AccessUndenied could not find a reason for AccessDenied.",
        access_denied_reason=common.AccessDeniedReason.ALLOWED,
    )
